// -----------------------------------------------------------------------
//   
//   Copyright 2003 H. Peter Anvin - All Rights Reserved
//
//   This program is free software; you can redistribute it and/or modify
//   it under the terms of the GNU General Public License as published by
//   the Free Software Foundation, Inc., 53 Temple Place Ste 330,
//   Bostom MA 02111-1307, USA; either version 2 of the License, or
//   (at your option) any later version; incorporated herein by reference.
//
// -----------------------------------------------------------------------
// "$Id: abc80.v,v 1.35 2003/10/26 18:52:22 hpa Exp $"

//
// abc80.v
//
// Top-level module for ABC80 FPGA implementation
//
// Designed for Altera Nios development kit, Cyclone Edition with
// the Lancelot add-on board from www.fpga.nl
//

// Define this to run at < 1 kHz and dump the fetch addresses
// to ttyb
// `define TTY_TRACE 1

// Select this to have the machine go to 40 characters after reset
// `define RESET_WIDTH 0
// Select this to have the machine go to 80 characters after reset
`define RESET_WIDTH 1

module abc80 (
              clkin,		// PLL1 input clock (50 MHz)
	      pld_clkout,	// Clock output to I/O devices
	      pld_clkfb,	// Feedback from pld_clkout to PLL2
              reset_n,		// Reset button
              ps2_sel,		// PS/2 port input/output select
              ps2_kclk,		// PS/2 keyboard clock
              ps2_kdata,	// PS/2 keyboard data
              ps2_mclk,		// PS/2 mouse clock
              ps2_mdata,	// PS/2 mouse data
	      led,		// Debugging LEDs
	      s7_0, s7_1,	// Debugging 7-segment LEDs
	      sw,		// Pushbutton switches
	      fse_a,		// Mainboard common bus address
	      fse_d,		// Mainboard common bus data
	      flash_cs_n,	// Flash ROM CS#
	      enet_aen,		// Ethernet Access Enable
	      sram_be_n,	// SRAM byte enables
	      sram_cs_n,	// SRAM CS#
	      sram_oe_n,	// SRAM OE#
	      sram_we_n,	// SRAM WE#
	      ttya_dcd,		// V.24 DCD#
	      ttya_txd,		// V.24 TxD
	      ttya_rxd,		// V.24 RxD
	      ttya_dtr,		// V.24 DTR#
	      ttya_dsr,		// V.24 DSR#
	      ttya_rts,		// V.24 RTS#
	      ttya_cts,		// V.24 CTS#
	      ttya_ri,		// V.24 RI#
	      ttyb_txd,		// Debug port TxD
	      vga_r,		// VGA red
	      vga_g,		// VGA green
	      vga_b,		// VGA blue
	      vga_hs,		// VGA horz sync
	      vga_vs, 		// VGA vert sync
	      vga_blank_n,	// VGA DAC force blank
	      vga_sync_n,	// VGA sync enable
	      vga_sync_t,	// VGA sync on R/G/B
	      vga_m1,		// VGA color space config
	      vga_m2,		// VGA color space config
	      cf_a,		// CompactFlash address bus
	      cf_d,		// CompactFlash data bus
	      cf_rdy,		// CompactFlash RDY
	      cf_wait_n,	// CompactFlash WAIT#
	      cf_ce1_n,		// CompactFlash CE1#
	      cf_ce2_n,		// CompactFlash CE2#
	      cf_oe_n,		// CompactFlash OE#
	      cf_we_n,		// CompactFlash WE#
	      cf_reg_n,		// CompactFlash REG#
	      cf_cd1_n		// CompactFlash card detect
	      );
   
   input         clkin;
   output 	 pld_clkout;
   input 	 pld_clkfb;
   input         reset_n;
   output        ps2_sel;
   inout         ps2_kclk;
   inout         ps2_kdata;
   inout         ps2_mclk;
   inout         ps2_mdata;
   input [3:0] 	 sw;
   output [7:0]  led;
   output [7:0]  s7_0;
   output [7:0]  s7_1;
   output [22:0] fse_a;
   inout [31:0]  fse_d;
   output 	 flash_cs_n;
   output 	 enet_aen;
   output [3:0]  sram_be_n;
   output 	 sram_cs_n;
   output 	 sram_oe_n;
   output 	 sram_we_n;
   output 	 ttya_dcd;	// DTE DTR
   output 	 ttya_txd;
   input 	 ttya_rxd;
   input 	 ttya_dtr;	// DTE DCD
   output 	 ttya_dsr;
   input 	 ttya_rts;	// DTE CTS
   output 	 ttya_cts;	// DTE RTS
   output 	 ttya_ri;
   output 	 ttyb_txd;
   output [7:0]  vga_r;
   output [7:0]  vga_g;
   output [7:0]  vga_b;
   output 	 vga_hs;
   output 	 vga_vs;
   output 	 vga_blank_n;
   output 	 vga_sync_n;
   output 	 vga_sync_t;
   output 	 vga_m1;
   output 	 vga_m2;
   output [10:0] cf_a;
   inout [15:0]  cf_d;
   input         cf_rdy;
   input         cf_wait_n;
   output        cf_ce1_n;
   output        cf_ce2_n;
   output        cf_oe_n;
   output        cf_we_n;
   output        cf_reg_n;
   input 	 cf_cd1_n;
   
   // ------------------------------------------------------------------------
   //  Reset
   // (there is option for extending the reset internally here, if necessary)
   // ------------------------------------------------------------------------
   wire 	 rst_n = reset_n;

   // ------------------------------------------------------------------------
   //  PLLs and clock distribution
   // ------------------------------------------------------------------------

   wire 	 cpu_clk;	//   25 MHz
   wire 	 video_clk;	//   25 MHz = screen pixel rate
   wire 	 fast_clk;	//  100 MHz
   wire 	 cpu_clk_d;

   pll1 pll1 (
	      .inclk0 ( clkin ), // 50 MHz input clock
	      .c0 ( fast_clk ),	 // x2/1 = 100 MHz output clock
	      .c1 ( cpu_clk_d ), // x1/2 =  25 MHz output clock
	      .e0 ( pld_clkout ) // External only output x1/2 = 25 MHz
	      );

   // pld_clkfb is pld_clkout routed externally through a zero-skew buffer
   pll2 pll2 (
	      .inclk0 ( pld_clkfb ), // 25 MHz input clock
	      .c0 ( video_clk )
	      );
   
`ifdef TTY_TRACE   
   reg [13:0] 	 cpu_clk_div;	// 763 Hz slow CPU clock for debugging
   always @(posedge cpu_clk_d)
     cpu_clk_div <= cpu_clk_div + 1;
   
   assign 	 cpu_clk = cpu_clk_div[13];
`else
   assign 	 cpu_clk = cpu_clk_d;
`endif
   
   // ------------------------------------------------------------------------
   //  Timers
   // ------------------------------------------------------------------------

   // ABC80 has two interrupt-capable timers: one at 128 us (7812.5 Hz) and
   // one at 20 ms (50 Hz).  This is assuming a 12 MHz crystal, which the real
   // ABC80 didn't have -- it had 11.9808 MHz, but everyone treated it as
   // 12 MHz (0.16% error).  We have a 50 MHz crystal; cpu_clk is already
   // divided by 2 to obtain 25 MHz (40 ns); we then divide it by 400 to
   // obtain a 16 us clock, which then is divided by 8 for 128 us and by
   // 625*2 for the 50 Hz clock.  Note the last stage is always divide by 2,
   // so we get 50% duty cycle.  I'm so anal-retentive.

   reg [8:0] 	 clk_div0;	// 16 us counter
   reg [2:0] 	 clk_div1;	// 128 us counter
   reg [9:0] 	 clk_div2;	// 10 ms counter
   reg 		 clk_div3;	// 20 ms counter
   
   wire		 ctr_128us;	// 128 us counter out
   wire		 ctr_20ms;	// 20 ms counter out

   always @(posedge cpu_clk)
     begin
	if ( clk_div0 == 0 )
	  begin
	     clk_div0 <= 9'd399;
	     
	     clk_div1 <= clk_div1 + 1; // No special wrapping needed
	     
	     if ( clk_div2 == 0 )
	       begin
		  clk_div2 <= 10'd624;
		  clk_div3 <= ~clk_div3;
	       end
	     else
	       clk_div2 <= clk_div2-1;
	  end // if ( clk_div0 == 0 )
	else
	  clk_div0 <= clk_div0-1;
     end // always @ (posedge cpu_clk)

   assign ctr_128us = clk_div1[2];
   assign ctr_20ms  = clk_div3;
	     
   // ------------------------------------------------------------------------
   //  Common bus wires declaration
   // ------------------------------------------------------------------------
   wire [15:0] 	 cpu_a;		// Output address from CPU (used for IORQ)
   wire [19:0] 	 mmu_a;		// Output address from MMU (within a device)
   reg [7:0] 	 msel;		// Decoded device selects for MREQ
   reg [7:0] 	 cpu_di;	// Input data to CPU
   wire [7:0] 	 cpu_do;	// Output data from CPU
   wire 	 cpu_m1_n;	// M1# from CPU
   wire 	 cpu_iorq_n;	// IORQ# from CPU
   wire 	 cpu_mreq_n;	// MREQ# from CPU
   wire 	 cpu_rd_n;	// RD# from CPU
   wire 	 cpu_wr_n;	// WR# from CPU
   wire 	 cpu_rfsh_n;	// RFSH# from CPU
   wire 	 cpu_halt_n;	// HALT# from CPU
   wire 	 cpu_busak_n;	// BUSAK# from CPU (not used)
   wire 	 cpu_int_n;	// INT# to CPU
   wire 	 cpu_nmi_n;	// NMI# to CPU
   // RETI# is a non-Z80 signal indicating that a RETI has been
   // executed.  We provide it in the CPU core because it is not feasible
   // for I/O devices to snoop the bus.
   wire 	 cpu_reti_n;	// RETI# from CPU
   
   // ABC-bus decoded strobes
   reg [7:0] 	 abc_out_n;
   reg [7:0] 	 abc_in_n;
   
   // ------------------------------------------------------------------------
   //  Keyboard controller
   // ------------------------------------------------------------------------

//`define FAKE_KEYBOARD 1

`ifdef FAKE_KEYBOARD
   wire [7:0] 	 kb_data = 8'h41;
   wire [7:0] 	 kb_stat;
   assign 	 kb_stat[6:0] = 0;
   assign 	 kb_stat[7] = ctr_128us;
`else
   wire [7:0] 	 kb_data;	// Keyboard symbol output
   wire [7:0] 	 kb_stat;	// Keyboard status word
   // Keyboard status word:
   // 7..0 = Strobe CapsLock NumLock ScrollLock Super Alt Ctrl Shift
   
   keyboard keyboard (
		      .clkin ( cpu_clk ),
		      .reset_n ( rst_n ),
		      .ps2_sel ( ps2_sel ),
		      .ps2_kclk ( ps2_kclk ),
		      .ps2_kdata ( ps2_kdata ),
		      .ps2_mclk ( ps2_mclk ),
		      .ps2_mdata ( ps2_mdata ),
		      .kb_data ( kb_data ),
		      .kb_stat ( kb_stat )
		      );
`endif
   
   // ------------------------------------------------------------------------
   //  Video controller
   // ------------------------------------------------------------------------

   wire [10:0] 	 video_a;
   wire [7:0] 	 video_d;
   wire [10:0] 	 chargen_a;
   wire [7:0] 	 chargen_d;
   wire [5:0] 	 rgb;
   wire [7:0] 	 vram_do;
   wire [7:0] 	 cgen_do;
   reg 		 video_width;
   
   videoram video_ram (
		       .wren_a ( 0 ),
		       .address_a ( video_a ),
		       .q_a ( video_d ),
		       .clock_a ( video_clk ),
		       .data_b ( cpu_do ),
		       .wren_b ( msel[2] & ~cpu_wr_n ),
		       .address_b ( mmu_a[11:0] ),
		       .q_b ( vram_do ),
		       .clock_b ( fast_clk )
		       );
   
   chargen char_gen (
		     .wren_a ( 0 ),
		     .address_a ( chargen_a ),
		     .q_a ( chargen_d ),
		     .clock_a ( video_clk ),
		     .data_b ( cpu_do ),
		     .wren_b ( msel[3] & ~cpu_wr_n ),
		     .address_b ( mmu_a[11:0] ),
		     .q_b ( cgen_do ),
		     .clock_b ( fast_clk )
		     );
      
   display video (
		  .clk ( video_clk ),
		  .width ( video_width ^ ~sw[1] ),
		  .reveal ( ~sw[0] ),
		  .a ( video_a ),
		  .d ( video_d ),
		  .ga ( chargen_a ),
		  .gd ( chargen_d ),
		  .rgb ( rgb ),
		  .vsync ( vga_vs ),
		  .hsync ( vga_hs )
		  );

   assign 	 vga_r = { rgb[5:4],rgb[5:4],rgb[5:4],rgb[5:4] };
   assign 	 vga_g = { rgb[3:2],rgb[3:2],rgb[3:2],rgb[3:2] };
   assign 	 vga_b = { rgb[1:0],rgb[1:0],rgb[1:0],rgb[1:0] };

   assign 	 vga_blank_n = 1;
   assign 	 vga_sync_n = 1;
   assign 	 vga_sync_t = 0; // Not sync on R/G/B
   assign 	 vga_m1 = 0;	// RGB color mode
   assign 	 vga_m2 = 0;	// RGB color mode
   
   // ------------------------------------------------------------------------
   //  External SRAM
   // ------------------------------------------------------------------------

   reg [7:0] 	 sram_do;	// Data out from sram
   wire		 sram_oe_w;
   wire		 sram_we_w;
   reg 		 sram_we_q1;
   reg 		 sram_we_q2;
   reg [3:0] 	 sram_be_q;
   
   assign flash_cs_n = 1;	// Disable flash ROM
   assign enet_aen   = 1;	// Disable Ethernet
   assign sram_cs_n  = 0;	// Select SRAM

   assign sram_oe_w = msel[0] & ~cpu_rd_n;
   assign sram_we_w = msel[0] & ~cpu_wr_n;

   always @(posedge fast_clk)
     begin
	// Delayed issues of WE# - to make sure the SRAM
	// doesn't latch data late
	sram_we_q1 <= sram_we_w;
	sram_we_q2 <= sram_we_q1;
     end // always @ (posedge fast_clk)

   // Driving output pins...
   // Note: The SRAM is connected to FSE_A[19:2] only...
   assign fse_a     = { 3'b000, mmu_a[19:2], 2'b00 };
   assign sram_be_n = 
	  (mmu_a[1:0] == 2'h0) ? ~4'b0001 :
	  (mmu_a[1:0] == 2'h1) ? ~4'b0010 :
	  (mmu_a[1:0] == 2'h2) ? ~4'b0100 :
	  /* (mmu_a[1:0] == 2'h3) ? */ ~4'b1000;

   assign sram_oe_n = ~sram_oe_w;
   assign sram_we_n = ~(sram_we_w & ~sram_we_q2);
   assign fse_d = sram_we_w ? { cpu_do, cpu_do, cpu_do, cpu_do } : 32'bz;
   
   // Input side
   always @(*)
     case ( mmu_a[1:0] )
       2'h0:
	 sram_do <= fse_d[7:0];
       2'h1:
	 sram_do <= fse_d[15:8];
       2'h2:
	 sram_do <= fse_d[23:16];
       2'h3:
	 sram_do <= fse_d[31:24];
     endcase // case( sram_a_q[1:0] )
   
   // ------------------------------------------------------------------------
   //  Internal ROMs
   // ------------------------------------------------------------------------

   wire [7:0]  brom_do;
   wire [15:0] brompatch;
   wire [7:0]  drom_do;

   // Full ROM; contains tag bit if patching needed in 80-column mode
   basicrom basic_rom (
		       .address( mmu_a[13:0] ),
		       .clock( fast_clk ),
		       .q ( brom_do )
		       );

   // Patch rom: contains alternate contents for 80-column mode
   basic80 basic_patch (
			.address( mmu_a[7:0] ),
			.clock( fast_clk ),
			.q ( brompatch )
			);

   // The patching itself is now done in the CPU data input
   
   // DOS ROM, nothing funny here...
   dosrom dos_rom (
		   .address( mmu_a[11:0] ),
		   .clock( fast_clk ),
		   .q ( drom_do )
		   );

`ifdef TTY_TRACE
   // Hold sw[1] to keep value from changing
   reg [7:0] 	 tty_clk_div;	// 115200 Hz tty out clock
   reg [50:0] 	 tty_data;
   wire 	 now_send;
   reg 		 prev_send;
   wire [7:0] 	 aa0;
   wire [7:0] 	 aa1;
   wire [7:0] 	 aa2;
   wire [7:0] 	 aa3;
   
   assign 	 now_send = ~cpu_mreq_n & ~cpu_m1_n;

   hexascii ha0 ( .value(cpu_a[3:0]),   .ascii(aa0) );
   hexascii ha1 ( .value(cpu_a[7:4]),   .ascii(aa1) );
   hexascii ha2 ( .value(cpu_a[11:8]),  .ascii(aa2) );
   hexascii ha3 ( .value(cpu_a[15:12]), .ascii(aa3) );

   always @(posedge video_clk)
     begin
	if ( tty_clk_div == 0 )
	  begin
	     tty_data <= { 1'b1, tty_data[50:1] };
	     tty_clk_div <= 8'd216; // tty_clk = video_clk/217
	  end
	else
	  tty_clk_div <= tty_clk_div - 1;

	if ( now_send & ~prev_send )
	  tty_data[50:1] <= { 1'b1, 8'h20, 1'b0,
			      1'b1, aa0, 1'b0,
			      1'b1, aa1, 1'b0,
			      1'b1, aa2, 1'b0,
			      1'b1, aa3, 1'b0 };

	prev_send <= now_send;
     end // always @ (posedge video_clk)

   assign ttyb_txd = tty_data[0];
 `endif
   
   // ------------------------------------------------------------------------
   //  MMU and memory address decoding
   // ------------------------------------------------------------------------

   wire [15:0]   mmu_q;		// The actual bits coming out of the MMU
   wire [2:0] 	 mmu_devsel;	// Undecoded device selects
   wire 	 mmu_patch;	// Patch BASIC?
   
   reg [9:0] 	 mmu_mod_addr;	// Modify address holding register
   reg [7:0] 	 mmu_mod_data;	// Modify data holding register (low byte)

   wire 	 mmu_wr_e;	// MMU write enable
   wire [15:0] 	 mmu_rd_data;	// Read data output from MMU

   reg [5:0] 	 mmu_map_sel;	// XXWWRR
   wire [1:0] 	 mmu_map;

   wire          mmu_sel;	// Selected for access via I/O ports
   reg [7:0] 	 mmu_do;	// Data out from MMU I/O ports
      
   assign 	 mmu_wr_e = mmu_sel & ~cpu_wr_n & (cpu_a[2:0] == 3'b011);
   assign 	 mmu_map =
		 ~cpu_m1_n ? mmu_map_sel[5:4] :
		 ~cpu_wr_n ? mmu_map_sel[3:2] :
		 mmu_map_sel[1:0];

   mmuram mmu (
	       .address_a ( { mmu_map, cpu_a[15:8] } ),
	       .data_a ( 16'bx ), // Never written
	       .q_a ( mmu_q ),
	       .wren_a ( 0 ),
	       .address_b ( mmu_mod_addr ),
	       .data_b ( { cpu_do, mmu_mod_data } ),
	       .q_b ( mmu_rd_data ),
	       .wren_b ( mmu_wr_e ),
	       .clock ( fast_clk )
	       );

   assign 	 mmu_patch = mmu_q[15];
   assign 	 mmu_devsel = mmu_q[14:12];
   assign 	 mmu_a[19:8] = mmu_q[11:0];
   assign 	 mmu_a[7:0] = cpu_a[7:0];

   always @(posedge fast_clk)
     begin
	if ( ~cpu_mreq_n )
	  case ( mmu_devsel )
	    3'h0:   msel <= 8'b00000001;
	    3'h1:   msel <= 8'b00000010;
	    3'h2:   msel <= 8'b00000100;
	    3'h3:   msel <= 8'b00001000;
	    3'h4:   msel <= 8'b00010000;
	    3'h5:   msel <= 8'b00100000;
	    3'h6:   msel <= 8'b01000000;
	    3'h7:   msel <= 8'b10000000;
	  endcase // case( mmu_devsel )
	else
	  msel <= 8'b00000000;
     end

   // Read and write MMU control registers
   always @(negedge rst_n or posedge fast_clk)
     begin
	if ( !rst_n )
	  begin
	     mmu_mod_addr <= 0;
	     mmu_mod_data <= 0;
	     mmu_map_sel  <= 0;
	     mmu_do <= ~8'b0;
	  end
	else
	  begin
	     mmu_do <= ~8'b0;
	     if ( mmu_sel & ~cpu_wr_n )
	       begin
		  case ( cpu_a[2:0] )
		    3'b000:
		      mmu_mod_addr[7:0] <= cpu_do[7:0];
		    3'b001:
		      mmu_mod_addr[9:8] <= cpu_do[1:0];
		    3'b010:
		      mmu_mod_data <= cpu_do[7:0];
		    3'b100:
		      mmu_map_sel <= cpu_do[5:0];
		  endcase // case( cpu_a[2:0] )
	       end // if ( mmu_sel & ~cpu_wr_n )
	     else if ( ~abc_out_n[7] )
	       begin
		  // For compatibility with Mikrodatorn's 64K hack
		  mmu_map_sel <= { cpu_do[1:0], cpu_do[1:0], cpu_do[1:0] };
	       end
	     else if ( mmu_sel & ~cpu_rd_n )
	       begin
		  case ( cpu_a[2:0] )
		    3'b000:
		      mmu_do <= mmu_mod_addr[7:0];
		    3'b001:
		      mmu_do <= { 6'b0, mmu_mod_addr[9:8] };
		    3'b010:
		      mmu_do <= mmu_rd_data[7:0];
		    3'b011:
		      mmu_do <= mmu_rd_data[15:8];
		    3'b100:
		      mmu_do <= { 2'b0, mmu_map_sel };
		  endcase // case( cpu_a[2:0] )
	       end // if ( mmu_sel & ~cpu_rd_n )
	  end
     end
   
   // ------------------------------------------------------------------------
   //  PIO
   // ------------------------------------------------------------------------

   reg [7:0] pioa_ai;
   wire [7:0] pioa_do;
   wire [7:0] piob_do;
   wire [7:0] piob_ao;
   wire       pio_sel;
   wire       pioa_ieo;
   wire       piob_ieo;
   wire       pioa_int_n;
   wire       piob_int_n;    

   always @(posedge cpu_clk)
     begin
	pioa_ai[6:0] <= kb_data[6:0]; // 7 bit ASCII
	pioa_ai[7]   <= kb_stat[7];   // Strobe
     end
   
   T80PIO pioa (
		.CLK_n ( cpu_clk ),
		.RESET_n ( rst_n ),
		.Di ( cpu_do ),
		.Do ( pioa_do ),
		.CE_n ( ~(pio_sel & ~cpu_a[1]) ),
		.CDsel ( cpu_a[0] ),
		.M1_n ( cpu_m1_n ),
		.IORQ_n ( cpu_iorq_n ),
		.RD_n ( cpu_rd_n ),
		.RETI_n ( cpu_reti_n ),
		.IEI ( 1 ),
		.IEO ( pioa_ieo ),
		.INT_n ( pioa_int_n ),
		.Ai ( pioa_ai ),
		.ASTB_n ( ctr_128us ),
		.BSTB_n ( /* == piob.astb_n */ 1 )
		);
   
   T80PIO piob (
		.CLK_n ( cpu_clk ),
		.RESET_n ( rst_n ),
		.Di ( cpu_do ),
		.Do ( piob_do ),
		.CE_n ( ~(pio_sel & cpu_a[1]) ),
		.CDsel ( cpu_a[0] ),
		.M1_n ( cpu_m1_n ),
		.IORQ_n ( cpu_iorq_n ),
		.RD_n ( cpu_rd_n ),
		.RETI_n ( cpu_reti_n ),
		.IEI ( pioa_ieo ),
		.IEO ( piob_ieo ),
		.INT_n ( piob_int_n ),
		.Ai ( { 5'b11111, ttya_dtr, ttya_rts, ttya_rxd } ),
		.Ao ( piob_ao ),
		.ASTB_n ( 1 ),
		.BSTB_n ( /* unused */ 1 )
		);

   assign     ttya_txd = piob_ao[3];
   assign     ttya_cts = piob_ao[4];
   assign     ttya_dcd = 0;
   assign     ttya_dsr = 0;
   assign     ttya_ri = 1;
   assign     led[5] = piob_ao[5]; // CAS relay
   assign     led[6] = piob_ao[6]; // CAS output
   assign     led[4:1] = 0;	// Unassigned ATM

   // ------------------------------------------------------------------------
   //  ABC-bus
   // ------------------------------------------------------------------------
   
   wire       abc_sel;
   
   always @(*)
     begin
	if ( abc_sel & ~cpu_wr_n )
	  case ( cpu_a[2:0] )
	    3'b000: abc_out_n = ~8'h01;
	    3'b001: abc_out_n = ~8'h02;
	    3'b010: abc_out_n = ~8'h04;
	    3'b011: abc_out_n = ~8'h08;
	    3'b100: abc_out_n = ~8'h10;
	    3'b101: abc_out_n = ~8'h20;
	    3'b110: abc_out_n = ~8'h40;
	    3'b111: abc_out_n = ~8'h80;
	  endcase // case( cpu_a[2:0] )
	else
	  abc_out_n = ~8'h00;
      
	if ( abc_sel & ~cpu_rd_n )
	  case ( cpu_a[2:0] )
	    3'b000: abc_in_n = ~8'h01;
	    3'b001: abc_in_n = ~8'h02;
	    3'b010: abc_in_n = ~8'h04;
	    3'b011: abc_in_n = ~8'h08;
	    3'b100: abc_in_n = ~8'h10;
	    3'b101: abc_in_n = ~8'h20;
	    3'b110: abc_in_n = ~8'h40;
	    3'b111: abc_in_n = ~8'h80;
	  endcase // case( cpu_a[2:0] )
	else
	  abc_in_n = ~8'h00;
     end // always @ (*)

   // CompactFlash controller
   wire [7:0] abc_cf_di;
   wire       cf_select;
   wire [8:0] cf_debug;

   cfcontroller cfcontroller (
			      .reset_n ( reset_n ),
			      .clk ( cpu_clk ),
			      .cf_a ( cf_a ),
			      .cf_d ( cf_d ),
			      .cf_rdy ( cf_rdy ),
			      .cf_wait_n ( cf_wait_n ),
			      .cf_ce1_n ( cf_ce1_n ),
			      .cf_ce2_n ( cf_ce2_n ),
			      .cf_oe_n ( cf_oe_n ),
			      .cf_we_n ( cf_we_n ),
			      .cf_reg_n ( cf_reg_n ),
			      .cf_cd1_n ( cf_cd1_n ),
			      .abc_do ( cpu_do ),
			      .abc_di ( abc_cf_di ),
			      .abc_out_n ( abc_out_n[0] ),
			      .abc_cs_n ( abc_out_n[1] ),
			      .abc_c1_n ( abc_out_n[2] ),
			      .abc_c2_n ( abc_out_n[3] ),
			      .abc_c3_n ( abc_out_n[4] ),
			      .abc_c4_n ( abc_out_n[5] ),
			      .abc_inp_n ( abc_in_n[0] ),
			      .abc_status_n ( abc_in_n[1] ),
			      .abc_rst_n ( abc_in_n[7] ),
			      .select ( cf_select ),
			      .debug ( cf_debug )
			      );

   // The terminology gets a bit funny there.  abc_do means data from
   // the ABC-bus to the main CPU.
   wire [7:0] abc_do = abc_cf_di;

   // Closest thing we get to a disk LED
   assign     led[0] = cf_select;

   // Video width control INP 4 = 80, INP 3 = 40
   always @(negedge reset_n or posedge cpu_clk)
     begin
	if ( ~reset_n )
	  video_width <= `RESET_WIDTH;
	else
	  if ( ~abc_in_n[3] )
	    video_width <= 0;
	  else if ( ~abc_in_n[4] )
	    video_width <= 1;
     end

   // Sound generator (not actually implemented, but flash the 7seg LEDs
   reg [7:0] sound;
   always @(negedge reset_n or posedge cpu_clk)
     begin
	if ( ~reset_n )
	  sound <= 0;
	else
	  if ( ~abc_out_n[6] )
	    sound <= cpu_do;
     end
      
   hexledx hexled0 (
		    .value( sound[3:0] ),
		    .blank( ~sound[0] ),
		    .minus( 0 ),
		    .s7 ( s7_0[6:0] )
		   );
   hexledx hexled1 (
		    .value( sound[7:4] ),
		    .blank( ~sound[0] ),
		    .minus( 0 ),
		    .s7 ( s7_1[6:0] )
		   );

   assign s7_0[7] = 1'b1;
   assign s7_1[7] = 1'b1;
   
   // ------------------------------------------------------------------------
   //  I/O address decoding
   // ------------------------------------------------------------------------

   wire [7:0] io_do;
   
   
   // In original ABC80, A5 selected between the PIO and the bus controller
   // (which included the sound generator.)  We use A7 for our own devices.
   // The PIO wants to see IORQ# and M1# directly.
   assign     pio_sel = ~cpu_a[7] & cpu_a[5];
   assign     abc_sel = ~cpu_iorq_n & cpu_m1_n & ~cpu_a[7] & ~cpu_a[5];
   assign     mmu_sel = ~cpu_iorq_n & cpu_m1_n & cpu_a[7] & ~cpu_a[4];  // Ports 80-87h
   wire       turbo_sel = ~cpu_iorq_n & cpu_m1_n & cpu_a[7] & cpu_a[4]; // Port 90h
   
   // All I/O devices emit all ones when not selected, so we just AND together
   // the outputs.
   assign     io_do = pioa_do & piob_do & mmu_do & abc_do;

   // ------------------------------------------------------------------------
   //  Main CPU
   // ------------------------------------------------------------------------

   assign     cpu_int_n = pioa_int_n & piob_int_n;
   assign     cpu_nmi_n = ctr_20ms;
   wire       cpu_wait_n = 1;	// Nothing to wait for atm
   wire       cpu_clk_en;
   reg 	      cpu_turbo;

   always @(*)
     begin
	if ( ~cpu_mreq_n & mmu_patch & video_width &
	     ( brompatch[15:8] == mmu_a[15:8] ) )
	  cpu_di <= brompatch[7:0];
	else
	  case ( msel )
	    8'b00000001: cpu_di <= sram_do;
	    8'b00000010: cpu_di <= 8'bx; // Currently unused
	    8'b00000100: cpu_di <= vram_do;
	    8'b00001000: cpu_di <= cgen_do;
	    8'b00010000: cpu_di <= brom_do;
	    8'b00100000: cpu_di <= drom_do;
	    8'b01000000: cpu_di <= 8'bx; // Currently unused
	    8'b10000000: cpu_di <= mmu_a[15:8];
	    8'b00000000: cpu_di <= io_do;
	    default:     cpu_di <= 8'bx;
	  endcase // case( msel )
     end // always @ (*)

   // Turbo control: sw2 = turbo sw3 = slow
   wire sw3_debounced;
   wire sw3_strobe;

   debounce sw3_debounce (
			 .clk ( cpu_clk ),
			 .reset_n ( rst_n ),
			 .in ( ~sw[3] ),
			 .out ( sw3_debounced ),
			 .strobe ( sw3_strobe )
			 );
   
   always @(negedge rst_n or posedge cpu_clk)
     begin
	if ( ~rst_n )
	  cpu_turbo <= 1'b1;
	else
	  begin
	     if ( turbo_sel & ~cpu_wr_n )
	       cpu_turbo <= cpu_do[0];
	     else if ( sw3_strobe & sw3_debounced )
	       cpu_turbo <= cpu_turbo ^ 1'b1;
	  end
     end

   // If cpu_turbo is 0, we want to execute at 3 MHz instead of 25
   // ... which means enabling the clock 3 times every 25 cycles
   reg [4:0] cpu_slow_ctr;
   always @(posedge cpu_clk)
     begin
	if ( cpu_slow_ctr[4:3] == 2'b11 )
	  cpu_slow_ctr <= 5'd0;
	else
	  cpu_slow_ctr <= cpu_slow_ctr + 1;
     end

   assign cpu_clk_en = cpu_turbo | (&cpu_slow_ctr[2:0]);
   
   assign led[7] = cpu_turbo;	// Turbo LED
   
   T80se cpu (
	      .clk_n ( cpu_clk ),
	      .clken ( cpu_clk_en ),
	      .reset_n ( rst_n ),
	      .wait_n ( cpu_wait_n ),
	      .int_n ( cpu_int_n ),
	      .nmi_n ( cpu_nmi_n ),
	      .busrq_n ( 1 ),
	      .busak_n ( cpu_busak_n ),
	      .reti_n ( cpu_reti_n ),
	      .m1_n ( cpu_m1_n ),
              .mreq_n ( cpu_mreq_n ),
              .iorq_n ( cpu_iorq_n ),
              .rd_n ( cpu_rd_n ),
              .wr_n ( cpu_wr_n ),
              .rfsh_n ( cpu_rfsh_n ),
              .halt_n ( cpu_halt_n ),
              .a ( cpu_a ),
              .di ( cpu_di ),
              .do ( cpu_do )
	      );
   
endmodule // abc80
